//Handwritten Number Classification for HackerBox 0095
//https://hackerboxes.com/products/hackerbox-0095-ai-camera
//
//Use of TensorFlow Lite Model is based on Trevor Lee's example:
//https://github.com/trevorwslee/Arduino-DumbDisplay/tree/master/projects/esp32tensorflowlite/esp32_mnist
//The mnist_model.h file can be obtained therefrom.
//
//GUI is based on on/off switch example from TFT_eSPI Library.

#include <TensorFlowLite_ESP32.h>
#include "tensorflow/lite/micro/all_ops_resolver.h"
#include "tensorflow/lite/micro/micro_error_reporter.h"
#include "tensorflow/lite/micro/micro_interpreter.h"
#include "tensorflow/lite/schema/schema_generated.h"
#include "mnist_model.h"

#include "FS.h"
#include <SPI.h>
#include <TFT_eSPI.h>

TFT_eSPI tft = TFT_eSPI();   

#define CALIBRATION_FILE "/TouchCalData3"

#define REPEAT_CAL false

#define INPUT_X 30
#define INPUT_Y 30
#define INPUT_W 120
#define INPUT_H 120

#define READBUTTON_X 210
#define READBUTTON_Y 50
#define READBUTTON_W 90
#define READBUTTON_H 50

#define CLEARBUTTON_X 210
#define CLEARBUTTON_Y 150
#define CLEARBUTTON_W 90
#define CLEARBUTTON_H 50


// error reporter for TensorFlow Lite
class tfErrorReporter : public tflite::ErrorReporter {
public:
  virtual int Report(const char* format, va_list args) {
    int len = strlen(format);
    char buffer[max(32, 2 * len)]; 
    sprintf(buffer, format, args);
    Serial.println(buffer);
    return 0;
  }
};

tflite::ErrorReporter* error_reporter = new tfErrorReporter();
const tflite::Model* model = ::tflite::GetModel(mnist_model_tflite);
const int tensor_arena_size = 81 * 1024;
uint8_t* tensor_arena;
tflite::MicroInterpreter* interpreter = NULL;
TfLiteTensor* input;

void setup() {
  Serial.begin(9600);
  if (model->version() != TFLITE_SCHEMA_VERSION) {
    Serial.println("Model not supported version");
  }

  // allocation memory for tensor_arena 
  tensor_arena = (uint8_t *) heap_caps_malloc(tensor_arena_size, MALLOC_CAP_INTERNAL | MALLOC_CAP_8BIT);
  if (tensor_arena == NULL) {
    Serial.println("heap_caps_malloc() failed");
    return;
  }

  // pull in all the operation implementations
  tflite::AllOpsResolver* resolver = new tflite::AllOpsResolver();

  // build an interpreter to run the model with
  interpreter = new tflite::MicroInterpreter(model, *resolver, tensor_arena, tensor_arena_size, error_reporter);

  // allocate memory from the tensor_arena for the model's tensors
  TfLiteStatus allocate_status = interpreter->AllocateTensors();
  if (allocate_status != kTfLiteOk) {
    Serial.println("AllocateTensors() failed");
    return;
  }

  // obtain a pointer to the model's input tensor
  input = interpreter->input(0);

  Serial.println("Done preparing MNIST TFLite model!");

  tft.init();
  tft.setRotation(3);
  touch_calibrate();
  tft.fillScreen(TFT_BLUE);
  ClearInput();

  PaintReadBtn();
  PaintClearBtn();
}

void loop()
{
  uint16_t x, y;

  if (tft.getTouch(&x, &y))
  {
    if ((x > READBUTTON_X) && (x < (READBUTTON_X + READBUTTON_W))) {
      if ((y > READBUTTON_Y) && (y <= (READBUTTON_Y + READBUTTON_H))) {
        Serial.println("Read button hit");
        ReadInputNumber();
      }
    }

    if ((x > CLEARBUTTON_X) && (x < (CLEARBUTTON_X + CLEARBUTTON_W))) {
      if ((y > CLEARBUTTON_Y) && (y <= (CLEARBUTTON_Y + CLEARBUTTON_H))) {
        Serial.println("Clear button hit");
        ClearInput();
      }
    }

    if ((x > INPUT_X + 8) && (x < (INPUT_X + INPUT_W - 8))) {
      if ((y > INPUT_Y + 8) && (y <= (INPUT_Y + INPUT_H - 8 ))) {
        tft.fillCircle(x, y, 2, TFT_BLACK);
        x=(x-INPUT_X)/4;  // compress 120 into 30
        y=(y-INPUT_Y)/4;  // compress 120 into 30
        input->data.f[(y*28)+x]=1;
        input->data.f[(y*28)+(x-1)]=1; //expand left one pixel
        input->data.f[(y*28)+(x+1)]=1; //expand right one pixel
        input->data.f[((y-1)*28)+x]=1; //expand up one pixel
        input->data.f[((y+1)*28)+x]=1; //expand down one pixel
      }
    }
  }
}

void touch_calibrate()
{
  uint16_t calData[5];
  uint8_t calDataOK = 0;

  // check file system exists
  if (!SPIFFS.begin()) {
    Serial.println("Formatting file system");
    SPIFFS.format();
    SPIFFS.begin();
  }

  // check if calibration file exists and size is correct
  if (SPIFFS.exists(CALIBRATION_FILE)) {
    if (REPEAT_CAL)
    {
      // Delete if we want to re-calibrate
      SPIFFS.remove(CALIBRATION_FILE);
    }
    else
    {
      File f = SPIFFS.open(CALIBRATION_FILE, "r");
      if (f) {
        if (f.readBytes((char *)calData, 14) == 14)
          calDataOK = 1;
        f.close();
      }
    }
  }

  if (calDataOK && !REPEAT_CAL) {
    // calibration data valid
    tft.setTouch(calData);
  } else {
    // data not valid so recalibrate
    tft.fillScreen(TFT_BLACK);
    tft.setCursor(20, 0);
    tft.setTextFont(2);
    tft.setTextSize(1);
    tft.setTextColor(TFT_WHITE, TFT_BLACK);

    tft.println("Touch corners as indicated");

    tft.setTextFont(1);
    tft.println();

    if (REPEAT_CAL) {
      tft.setTextColor(TFT_RED, TFT_BLACK);
      tft.println("Set REPEAT_CAL to false to stop this running again!");
    }

    tft.calibrateTouch(calData, TFT_MAGENTA, TFT_BLACK, 15);

    tft.setTextColor(TFT_GREEN, TFT_BLACK);
    tft.println("Calibration complete!");

    // store data
    File f = SPIFFS.open(CALIBRATION_FILE, "w");
    if (f) {
      f.write((const unsigned char *)calData, 14);
      f.close();
    }
  }
}

void PaintReadBtn()
{
  tft.fillRect(READBUTTON_X, READBUTTON_Y, READBUTTON_W, READBUTTON_H, TFT_RED);
  tft.setTextColor(TFT_WHITE);
  tft.setTextSize(2);
  tft.setTextDatum(MC_DATUM);
  tft.drawString("Read", READBUTTON_X + (READBUTTON_W / 2), READBUTTON_Y + (READBUTTON_H / 2));
}

void PaintClearBtn()
{
  tft.fillRect(CLEARBUTTON_X, CLEARBUTTON_Y, CLEARBUTTON_W, CLEARBUTTON_H, TFT_RED);
  tft.setTextColor(TFT_WHITE);
  tft.setTextSize(2);
  tft.setTextDatum(MC_DATUM);
  tft.drawString("Clear", CLEARBUTTON_X + (CLEARBUTTON_W / 2) + 1, CLEARBUTTON_Y + (CLEARBUTTON_H / 2));
}

void ClearInput()
{
  for (int i=0; i<28*28; i++)
    input->data.f[i]=0;
  tft.fillRect(INPUT_X, INPUT_Y, INPUT_W, INPUT_H, TFT_WHITE);
  tft.fillRect(80, 185, 50, 80, TFT_BLUE);
}

void ReadInputNumber()
{
    // run the model on the input pixels
    Serial.println("Invoke Model");
    long detect_start_millis = millis();
    TfLiteStatus invoke_status = interpreter->Invoke();
    if (invoke_status != kTfLiteOk) {
      Serial.println("Model Failed");
    }

    // process the inference results
    TfLiteTensor* output = interpreter->output(0);
    if (true) {
      for (int i = 0; i < 10; i++) {
        float p = output->data.f[i];
        Serial.println(String(". ") + String(i) + ": " + String(p, 3));
      }
    }

    int best = -1;
    float bestProp;
    for (int i = 0; i < 10; i++) {
      float prop = output->data.f[i];
      if (i == 0) {
        best = 0;
        bestProp = prop; 
      } else if (prop > bestProp) {
        best = i;
        bestProp = prop;
      }
    }
    Serial.println(String("*** [") + best + "] (" + bestProp + ") " );  
    tft.setCursor(80, 185);
    tft.setTextFont(2);
    tft.setTextSize(2);
    tft.setTextColor(TFT_BLACK, TFT_WHITE);
    tft.println(best);
}

